import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Stroke;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * Turtle window class. This class implements a window in which turtles can draw.
 *
 * @author malcolmr
 * @version 16 July 2012
 */
public class Window implements ChangeListener {

    private List<Turtle> myTurtles;
    private BufferedImage myImage;

    // the turtle polygon is a simple triangle
    private static final int[] XPOINTS = { 10, -5, -5 };
    private static final int[] YPOINTS = { 0, 5, -5 };
    private static final Polygon TURTLE_POLYGON = new Polygon(XPOINTS, YPOINTS,
            XPOINTS.length);

    private static final Dimension DEFAULT_SIZE = new Dimension(800, 800);
    private Dimension mySize;
    private JFrame myFrame;
    private TurtlePanel myPanel;
    private JCheckBox myShowTurtlesButton;

    /**
     * A private JPanel class which draws an image and a bunch of turtles
     */
    
    private class TurtlePanel extends JPanel {
        private static final long serialVersionUID = 1L;

        /**
         * Repaint the image and the turtles
         */
        public void paint(Graphics g) {
            super.paint(g);
            Graphics2D g2 = (Graphics2D) g;

            // paint the image
            if (myImage != null) {
                g.drawImage(myImage, 0, 0, this);
            }

            // paint the turtles if they are visible
            if (visibleTurtles()) {
                // place the origin at the centre of the window
                g2.translate(mySize.width / 2, mySize.height / 2);

                for (Turtle t : myTurtles) {
                    Point p = t.position();
                    g2.translate(p.x, p.y);

                    // rotate each turtle to face its given angle 
                    double theta = Math.toRadians(t.rotation());
                    g2.rotate(theta);
                    drawTurtle(g2);

                    // undo the transformation for the next turtle
                    g2.rotate(-theta);
                    g2.translate(-p.x, -p.y);
                }
            }
        }

        /**
         * Draw an individual turtle.
         */
        private void drawTurtle(Graphics2D g2) {
            g2.setColor(Color.GREEN);
            g2.fillPolygon(TURTLE_POLYGON); // the interior
            g2.setColor(Color.BLACK);
            g2.drawPolygon(TURTLE_POLYGON); // the outline
            g2.drawOval(-1, -1, 2, 2);  // the pen marker
        }

    }

    /**
     * Create a new window of default size.
     */
    public Window() {
        this("World", DEFAULT_SIZE.width, DEFAULT_SIZE.height);
    }

    /**
     * Create a new window with the specified name and dimensions.
     */
    public Window(String name, int width, int height) {
        mySize = new Dimension(width, height);
        
        // Create an image to draw on. Clear it to white.
        myImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
        Graphics2D g = myImage.createGraphics();
        g.setBackground(Color.WHITE);
        g.clearRect(0, 0, width, height);
        myTurtles = new ArrayList<Turtle>();

        // create a frame (window)
        myFrame = new JFrame(name);
        myFrame.setLayout(new BorderLayout());

        // create a turtle panel to display the image and turtles
        myPanel = new TurtlePanel();
        myPanel.setPreferredSize(new Dimension(width, height));
        myFrame.add(myPanel, BorderLayout.CENTER);

        // a button to toggle whether the turtles are visible
        myShowTurtlesButton = new JCheckBox("Show turtles");
        myShowTurtlesButton.setSelected(true);
        myShowTurtlesButton.addChangeListener(this);
        myFrame.add(myShowTurtlesButton, BorderLayout.PAGE_END);

        myFrame.setResizable(false);
        myFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        myFrame.pack();
        myFrame.setVisible(true);
    }

    /**
     * From the ChangeListener interface. Respond to 'change' events.
     */
    public void stateChanged() {
        myPanel.repaint();
    }

    /**
     * From the ChangeListener interface. Respond to 'change' events.
     */
    @Override
    public void stateChanged(ChangeEvent e) {
        myPanel.repaint();
    }

    /**
     * Check if the turtles are visible.
     */
    public boolean visibleTurtles() {
        return myShowTurtlesButton.isSelected();
    }

    /**
     * Set the visibility of the turtles.
     */
    public void setVisibleTurtles(boolean visible) {
        myShowTurtlesButton.setSelected(visible);
    }

    /**
     * Add another turtle to the window.
     */
    public void addTurtle(Turtle t) {
        myTurtles.add(t);
        myPanel.repaint();
    }

    /**
     * Get a list of turtles in this window. Note that this list is unmodifiable.
     */
    public List<Turtle> turtles() {
        return Collections.unmodifiableList(myTurtles);
    }

    /**
     * Draw a line on the background image.
     */
    public void drawLine(int x1, int y1, int x2, int y2, Color color,
            Stroke stroke) {

        Graphics2D g2 = myImage.createGraphics();
        g2.setColor(color);
        g2.setStroke(stroke);
        g2.translate(myImage.getWidth() / 2, myImage.getHeight() / 2);
        g2.drawLine(x1, y1, x2, y2);
        myPanel.repaint();
    }

    /**
     * Sample main program. Draws a spiral.
     */
    public static void main(String[] args) {
        Window w = new Window();
        Turtle t = new Turtle(w);

        t.setPenWidth(5);
        t.setPenColor(new Color(Color.HSBtoRGB(0, 1, 1)));

        for (int i = 0; i < 100; i++) {
            t.move(i);
            t.rotate(45);
        }

    }

}
